/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * DirectionConverter.hpp
 *
 *  Created on: 2011/03/10
 *      Author: HARK Development Team
 */

#ifndef DIRECTIONCONVERTER_HPP_
#define DIRECTIONCONVERTER_HPP_

/**
 * @file DirectionConverter.hpp
 * HARK標準座標系の角度と、hark-binaural内部で用いる角度の相互変換を提供する.
 */

#include <cstddef> // size_t

namespace NSHarkBinaural
{

/**
 * HARK標準座標系とhark-binaural内部の角度インデックス番号変換器.
 * 隣り合うインデックス間の角度が一定の場合にのみ対応している。
 * インデックスiの角度をdir(i)とすると、どのようなiに対しても
 *   dir(i + 1) - dir(i) == dir(i + 2) - dir(i + 1)
 * となるような場合のみ対応しているということ。
 */
class DirectionConverter
{
public:
    /**
     * 変換器を生成する.
     * @param increaseDegree インデックスの1増加に対応する度数の増加量（インデックス増加に応じて角度が減少するなら負値）
     * @param beginDegree インデックス0に対応する度数
     */
    DirectionConverter(int increaseDegree, int beginDegree);

    /**
     * HARK標準座標系の度数から角度インデックス番号へ変換する.
     * degree - beginDegreeがincreaseDegreeの倍数でない場合の動作は定義されない。
     * @param degree 度数
     * @return インデックス番号
     */
    size_t ToIndex(int degree) const;

    /**
     * 角度インデックス番号からHARK標準座標系の度数へ変換する.
     */
    int ToDegree(size_t index) const;

    /**
     * インデックスの1増加に対応する度数の増加量を取得する.
     */
    int IncreaseDegree() const {
        return this->increaseDegree;
    }

    /**
     * インデックス0に対応する度数を取得する.
     */
    int BeginDegree() const {
        return this->beginDegree;
    }

private:
    /// インデックスの1増加に対応する度数の増加量.
    int increaseDegree;

    /// インデックス0に対応する度数.
    int beginDegree;
};

/**
 * グローバル角度変換器.
 * 特に理由がない限り、hark-binaural内部ではこの変換器を使って角度を扱えばよい。
 */
extern DirectionConverter gDirConv;

}

#endif /* DIRECTIONCONVERTER_HPP_ */
