/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * Harmonics.cpp
 *
 *  Created on: 2010/12/10
 *      Author: HARK Development Team
 */

#include <vector>
#include <stdexcept>
#include <boost/shared_ptr.hpp>
#include <DimensionPromise/types/Frequency.hpp>
#include <DimensionPromise/operators.hpp>
#include <HarmonicSound.hpp>
#include <Harmonics.hpp>
#include <Logger.hpp>
#include <boolean.hpp>

using namespace std;
using namespace boost;

namespace NSHarkBinaural
{

Harmonics::HarmonicSoundVectorPtr
Harmonics::ExtractAllHarmonics(const BandVectorPtr proxyBands)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);
    if (proxyBands.get() == 0)
    {
        LogWarning(logger, "proxyBands is null");
        throw invalid_argument("proxyBands is null");
    }
    // 使いやすくするための参照はずし
    const BandVector& bands = *proxyBands;

    HarmonicSoundPtr sound;

    HarmonicSoundVectorPtr result(new HarmonicSoundVector());
    HarmonicSoundVector& sounds = *result;

    for (size_t i = 0; i < bands.size(); ++i)
    {
        sound = GetHarmonics(proxyBands, i);
        if (sound.get() != 0)
        {
            // 調波構造を発見した
            sounds.push_back(sound);
        }
    }

    LogFine(logger, format("%d harmonic structures were found") % sounds.size());
    return result;
}

Harmonics::HarmonicSoundVectorPtr
Harmonics::GetHumanHarmonics(const HarmonicSoundVectorPtr harmonicsList, size_t bandSize, const BandInformation& bi, int threshold)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);
    if (harmonicsList.get() == 0)
    {
        LogWarning(logger, "harmonicsList is null");
        throw invalid_argument("harmonicsList is null");
    }
    // 使いやすくするための参照はずし
    const HarmonicSoundVector& harmonicsList_ = *harmonicsList;

    HarmonicSoundPtr sound;

    HarmonicSoundVectorPtr result(new HarmonicSoundVector());
    HarmonicSoundVector& sounds = *result;

    // 使用した周波数バンドの一覧（最初はどのバンドも使っていないので全要素false）
    vector<boolean> used(bandSize, false);

    int check;
    for (HarmonicSoundVector::const_iterator srcSound = harmonicsList_.begin(); srcSound != harmonicsList_.end(); ++srcSound)
    {
        check = CheckHarmonics(*srcSound, bi);
#ifdef S500_600
        if( check > -1 )
#else
        if (check >= threshold)
#endif
        {
            // used 更新
            const HarmonicSoundPtr& sound = *srcSound;
            HarmonicSound& sound_ = *sound;
            if (used.at(sound_.bands->at(0).band.Index()))
            {
                // sound_の基音が既に違う調波構造の倍音として現れている場合は
                // sound_を調波構造としてカウントしない
                continue;
            }
            for (vector<BandAndHarmonicRatio>::const_iterator it = sound_.bands->begin(); it != sound_.bands->end(); ++it)
            {
                used.at(it->band.Index()) = true;
            }
            sounds.push_back(sound);
        }
    }

    LogFine(logger, format("%d human harmonic structures were found") % sounds.size());
    return result;
}

Band
Harmonics::GetNHarmonic(const BandVectorPtr proxyBands, int fundamental, int n)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);
    int nIndex; // 計算によって求めたn倍音の周波数バンドインデックス
    int approximatedIndex; // proxyBandsの中から選ばれたnIndexの近似値
    int diff; // nIndexとapproximatedIndexの差

    if (proxyBands.get() == 0)
    {
        LogWarning(logger, "proxyBands is null");
        throw invalid_argument("proxyBands is null");
    }
    if (n <= 1)
    {
        return Band();
    }
    if (proxyBands->size() <= fundamental)
    {
        LogFine(logger, "proxyBands' size <= fundamental");
        return Band();
    }
    if (!proxyBands->at(fundamental).IsValid())
    {
        LogWarning(logger, "given invalid band");
        throw invalid_argument("given invalid band");
    }

    // 使いやすくするための参照はずし
    BandVector& bands = *proxyBands;

    /* N倍音のインデックス */
    nIndex = bands.at(fundamental).Index() * n;

    for (size_t i = 0; i < bands.size(); ++i)
    {
        if (bands[i].Index() > nIndex)
        {
            // 差が小さい方を採用する。
            Band& approximatedBand = ((nIndex - bands[i - 1].Index()) > (bands[i].Index() - nIndex) ?
                                      bands[i] : bands[i - 1]);
            approximatedIndex = approximatedBand.Index();
            diff = nIndex - approximatedIndex;

            switch (bands[fundamental].Index()) // TODO なぜ、基音によって場合分けしているのか？nによって場合分けではないのだろうか？
            {
            case 1: // TODO なぜ代表値1に対しての倍音は取得できないのか？（n == 1のときに倍音無しとするのならば意味が分かる）
                return Band();
                break;
            case 2:
                return diff == 0 ? approximatedBand : Band();
            case 3:
                return abs(diff) < 1 ? approximatedBand : Band();
            default:
                if (bands[fundamental].Index() >= 4)
                {
                    return abs(diff) < 2 ? approximatedBand : Band();
                }
                else
                {
                    return Band();
                }
            }
        }
    }
    return Band();
}

Harmonics::HarmonicSoundPtr
Harmonics::GetHarmonics(const BandVectorPtr proxyBands, int fundamental)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);

    if (proxyBands.get() == 0)
    {
        LogWarning(logger, "proxyBands is null");
        throw invalid_argument("proxyBands is null");
    }
    if (proxyBands->size() <= fundamental)
    {
        LogFine(logger, "proxyBands' size <= fundamental");
        return HarmonicSoundPtr();
    }

    HarmonicSoundPtr harmonics;

    LogFine(logger, format("searching 2 harmonic (fundamental band = %d)") % proxyBands->at(fundamental).Index());
    Band n2Harmonic = GetNHarmonic(proxyBands, fundamental, 2);
    if (n2Harmonic.IsValid())
    {
        // 2倍音がある。
        LogFine(logger, format("there is 2 harmonic: 2 harmonic band = %d") % n2Harmonic.Index());
        harmonics = HarmonicSoundPtr(new HarmonicSound());
        HarmonicSound& sound = *harmonics;
        sound.bands->push_back(BandAndHarmonicRatio(proxyBands->at(fundamental), 1));
        sound.bands->push_back(BandAndHarmonicRatio(n2Harmonic, 2));

        // 最高の倍音の、基音に対する倍率を得る
        int nMax = proxyBands->at(fundamental).Index() == 0 ? proxyBands->back().Index() :
                   static_cast<int>(
                       static_cast<double>(proxyBands->back().Index()) /
                       static_cast<double>(proxyBands->at(fundamental).Index()) + 0.5);
        for (int n = 3; n <= nMax; ++n)
        {
            // n倍音を得る
            Band nHarmonic = GetNHarmonic(proxyBands, fundamental, n);
            // n倍音が存在する場合のみ追加する
            if (nHarmonic.IsValid())
            {
                LogFinest(logger, format("there is %d harmonic: %d harmonic band = %d") % n % n % nHarmonic.Index());
                sound.bands->push_back(BandAndHarmonicRatio(nHarmonic, n));
            }
        }
    }
#ifdef S500_600
    /* pure tone 用 */
    else
    {
        sound.push_back(BandAndHarmonicRatio(proxyBands->at(fundamental), 1));
    }
#endif

    return harmonics;
}

#if 1
int
Harmonics::CheckHarmonics(const HarmonicSoundPtr sound, const BandInformation& bi)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);
    if (sound.get() == 0)
    {
        LogWarning(logger, "sound is null");
        throw invalid_argument("sound is null");
    }
    int flag = -4, harmonicsCount = 0;
    int h5;
    const BandAndHarmonicRatio& highestHarmonics = sound->bands->back();
    // 5倍音まで調べる
    h5 = highestHarmonics.ratio <= 5 ? highestHarmonics.ratio : 5;

    LogFine(logger, format("cheking harmonics, highest harmonics ratio = %d") % highestHarmonics.ratio);

    // 存在する倍音
    vector<boolean> existRatio(highestHarmonics.ratio, false);
    for (int i = 0; i < sound->bands->size(); ++i)
    {
        existRatio.at(sound->bands->operator [](i).ratio - 1) = true;
        if (sound->bands->operator [](i).ratio <= h5) harmonicsCount++;
    }

    LogFine(logger, format("bands size = %d, harmonicsCount = %d") % sound->bands->size() % harmonicsCount);
    if (harmonicsCount >= 3) flag = 4;

    if (highestHarmonics.ratio <= 5)
    {
        flag = -4;
        if (sound->bands->size() - harmonicsCount <= 1)
            flag = 2;
        if (sound->bands->size() == 1 && harmonicsCount == 1)
            flag = 1; /* pure tone */
    }

    /* 基音が300Hz以上なら -> 人の声ではない */
    if (sound->bands->operator [](0).band.Freq() > NSDimensionPromise::Frequency::FromHz(300))
        flag = 0;
#ifdef S500_600
    if( sound->harms[0] > (int)(605.0 / SAMPLINGRATE * NFFT) ) flag = -10;
    if( sound->harms[0] < (int)(595.0 / SAMPLINGRATE * NFFT) &&
            sound->harms[0] > (int)(505.0 / SAMPLINGRATE * NFFT) ) flag = -10;
    if( sound->harms[0] < (int)(495.0 / SAMPLINGRATE * NFFT) ) flag = -100;
#endif
    /*
    if (sound->harms[0] == 0 && sound->harms[2] == 0) // 基音が0Hz付近で、3倍音無し
      flag = -2;
    if (sound->harms[0] == 0 && sound->harms[1] == 0) // 基音が0Hz付近で、2倍音無し
      flag = -3;
    */
    if (sound->bands->operator [](0).band.Index() == 0 && !existRatio.at(2))
        flag = -2;
    if (sound->bands->operator [](0).band.Index() == 0 && !existRatio.at(1))
        flag = -3;

    if (sound->bands->size() >= 7)
    {
        // 3倍音、5倍音、7倍音が無し
        //if (sound->harms[2] == 0 && sound->harms[4] == 0 && sound->harms[6] == 0)
        if (!existRatio.at(2) && !existRatio.at(4) && !existRatio.at(6))
            flag = -4;
    }
    return flag;
}
#else

/**
 * 内田オリジナル.
 */
int
Harmonics::CheckHarmonics(const HarmonicSoundPtr sound, const BandInformation& bi)
{
    static Logger& logger = Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME);
    if (sound.get() == 0)
    {
        LogWarning(logger, "sound is null");
        throw invalid_argument("sound is null");
    }
    int flag = -4, harmonicsCount = 0;
    int h5;
    const BandAndHarmonicRatio& highestHarmonics = sound->bands->back();
    const int maxHarm = 10;
    // maxHarm倍音まで調べる
    h5 = max(highestHarmonics.ratio, maxHarm);

    LogFine(logger, format("cheking harmonics, highest harmonics ratio = %d") % highestHarmonics.ratio);

    // 存在する倍音
    vector<boolean> existRatio(highestHarmonics.ratio, false);
    for (int i = 0; i < sound->bands->size(); ++i)
    {
        existRatio.at(sound->bands->operator [](i).ratio - 1) = true;
        if (sound->bands->at(i).ratio <= h5) harmonicsCount++;
    }

    LogFine(logger, format("bands size = %d, harmonicsCount = %d") % sound->bands->size() % harmonicsCount);
    flag = harmonicsCount;

    if (highestHarmonics.ratio <= 5)
    {
        flag = -4;
        if (sound->bands->size() - harmonicsCount <= 1)
            flag = 2;
        if (sound->bands->size() == 1 && harmonicsCount == 1)
            flag = 1; /* pure tone */
    }

    /* 基音が400Hz以上なら -> 人の声ではない */
    if (sound->bands->operator [](0).band.Freq() > NSDimensionPromise::Frequency::FromHz(400))
        flag = 0;
#ifdef S500_600
    if( sound->harms[0] > (int)(605.0 / SAMPLINGRATE * NFFT) ) flag = -10;
    if( sound->harms[0] < (int)(595.0 / SAMPLINGRATE * NFFT) &&
            sound->harms[0] > (int)(505.0 / SAMPLINGRATE * NFFT) ) flag = -10;
    if( sound->harms[0] < (int)(495.0 / SAMPLINGRATE * NFFT) ) flag = -100;
#endif
    /*
    if (sound->harms[0] == 0 && sound->harms[2] == 0) // 基音が0Hz付近で、3倍音無し
      flag = -2;
    if (sound->harms[0] == 0 && sound->harms[1] == 0) // 基音が0Hz付近で、2倍音無し
      flag = -3;
    */
    if (sound->bands->operator [](0).band.Index() == 0 && !existRatio.at(2))
        flag = -2;
    if (sound->bands->operator [](0).band.Index() == 0 && !existRatio.at(1))
        flag = -3;

    if (sound->bands->size() >= 7)
    {
        // 3倍音、5倍音、7倍音が無し
        //if (sound->harms[2] == 0 && sound->harms[4] == 0 && sound->harms[6] == 0)
        if (!existRatio.at(2) && !existRatio.at(4) && !existRatio.at(6))
            flag = -4;
    }
    return flag;
}

#endif

}
